
/* 
 * File: "CoxeterExteions_by_EM_InvolPoset.h"
 *
 * Author: Eric Marberg
 * Version: June 2013
 * --------------------------------------------------------------
 * Header file for the TwistedInvolutionsPoset class. This is the main
 * class in our batch of Coxeter extensions; it implements the Bruhat
 * ordered poset of twisted involutions I_* in a finite Coxeter group
 * (W,S) with S-preserving involution *. 
 *
 */



#ifndef __TwistedInvolutionsPoset__
#define __TwistedInvolutionsPoset__



#include "commands.h"
#include "interactive.h"
#include "fcoxgroup.h"

#include "CoxeterExtensions_by_EM_SafePolynomial.h"
#include "CoxeterExtensions_by_EM_InvolBruhatIterator.h"
#include "CoxeterExtensions_by_EM_Structs.h"

using namespace std;
using namespace commands;



/*
 * Constants
 * ---------------------------------------------------------
 * The programs creats a number of output files which are stored
 * in an output directory. This directory must exist before running
 * any calculations, and should be structured as follows:
 *
 *   BASE_FOLDER
 *     DATA_FOLDER
 *       POLYNOMIALS_P_FOLDER
 *       POLYNOMIALS_Psigma_FOLDER
 *       STRUCTURECONSTANTS_h_FOLDER
 *       STRUCTURECONSTANTS_htilde_FOLDER
 *		 STRUCTURECONSTANTS_hsigma_FOLDER
 *       INDICES_ALL_ELEMENTS_FOLDER
 *       INDICES_INVOLUTIONS_FOLDER
 *     OUTPUT_FOLDER
 *
 * The macros below indicate the user defined names of these folders
 * (and of the output files to be created in them).
 *
 */

#define BASE_FOLDER			"CoxeterExtensions_by_EM_Output/"

#define DATA_FOLDER			BASE_FOLDER "Data/"
#define OUTPUT_FOLDER		BASE_FOLDER "Output/"

#define POLYNOMIALS_P_LABEL						"Polynomials_P"
#define POLYNOMIALS_Psigma_LABEL				"Polynomials_Psigma"
#define STRUCTURECONSTANTS_h_LABEL				"StructureConstants_h"
#define STRUCTURECONSTANTS_htilde_LABEL			"StructureConstants_htilde"
#define STRUCTURECONSTANTS_hsigma_LABEL			"StructureConstants_hsigma"
#define INDICES_ALL_ELEMENTS_LABEL				"Indices_AllElements"
#define INDICES_INVOLUTIONS_LABEL				"Indices_Involutions"
#define Summary_LABEL							"Summary"

#define STRUCTURECONSTANTS_h_TEMP_LABEL			STRUCTURECONSTANTS_h_LABEL		"_yindex"
#define STRUCTURECONSTANTS_htilde_TEMP_LABEL	STRUCTURECONSTANTS_htilde_LABEL "_xindex"

#define POLYNOMIALS_P_FOLDER					DATA_FOLDER POLYNOMIALS_P_LABEL				"/"
#define POLYNOMIALS_Psigma_FOLDER				DATA_FOLDER POLYNOMIALS_Psigma_LABEL		"/"
#define STRUCTURECONSTANTS_h_FOLDER				DATA_FOLDER STRUCTURECONSTANTS_h_LABEL		"/"
#define STRUCTURECONSTANTS_htilde_FOLDER		DATA_FOLDER STRUCTURECONSTANTS_htilde_LABEL "/"
#define STRUCTURECONSTANTS_hsigma_FOLDER		DATA_FOLDER STRUCTURECONSTANTS_hsigma_LABEL "/"
#define INDICES_ALL_ELEMENTS_FOLDER				DATA_FOLDER INDICES_ALL_ELEMENTS_LABEL		"/"
#define INDICES_INVOLUTIONS_FOLDER				DATA_FOLDER INDICES_INVOLUTIONS_LABEL		"/"


// The following constants indicate the name of folders used to store some 
// temporary files which the program creates and (ultimately) deletes. At 
// present we just reuse two already existing folders for this purpose.

#define STRUCTURECONSTANTS_h_TEMP_FOLDER		STRUCTURECONSTANTS_h_FOLDER
#define STRUCTURECONSTANTS_htilde_TEMP_FOLDER	STRUCTURECONSTANTS_htilde_FOLDER



typedef std::map<int,SafePolynomial> InvolIndexPol_Map;


/*
 * Class: TwistedInvolutionsPoset
 * ---------------------------------------------------------
 * This class implements various computations involving the set of
 * twisted involutions in a finite Coxeter group and the Bruhat ordering
 * on this set.
 *
 * The constructor queries the user for the type and rank of a finite 
 * Coxeter group (W,S), and also for an S-preserving involution of W.
 * The Bruhat order on the set of twisted involutions I_* is then computed.
 * 
 * The main functions called on the TwistedInvolutionsPoset object are the
 * following void methods with no arguments:
 *
 *   Compute_P_Polynomials
 *	 Compute_Psigma_Polynomials
 *   Compute_hsigma_StructureConstants
 *   Test_Polynomials
 *   Test_StructureConstants
 *
 * The first three of these functions compute the polynomials P_{y,w},
 * Psigma_{y,w}, and hsigma_{x,y;z} attached to (W,S,*). The last two
 * functions test whether the polynomials P^{+/-}_{y,w} and h^{+/-}_{x,y;z}
 * possess various conjectured positivity properties.
 * 
 */
class TwistedInvolutionsPoset {
	
	interactive::CoxGroup* W;
	
	// This list of Generators represents the involuion *
	CoxWord star;
	
	string group_name;
	bool verbose_printing_on;
	
	std::vector<CoxNbr> involutions;	
	
	// elem_actions[w][s] gives the index of the twisted involution
	// sw or sws*, whichever is not equal to w and in I_*
	//
	int** elem_actions;
	
	// elem_lengths[w] gives the lenght in W of the twisted involution (indexed by) w
	int* elem_lengths;
	
	// bruhat_order[y][w] is true if and only if the twisted involution
	// indexed by y is <= the twisted involution indexed by w
	bool** bruhat_order;
	
	// bruhat_ascents[y] gives the set of (indices) of twisted
	// involutions z such that z covers y in the Bruhat order on I_*
	//
	std::vector<int>* bruhat_ascents;
	
	// bruhat_ascents_by_descents[s][w][y] gives the set of 
	// (indices of) twisted involutions z such that z covers y in
	// the Bruhat order on I_* and z <= w and s is a left descent of z.
	//
	std::vector<int>*** bruhat_ascents_by_descents;

	// Struct which stores data associated with Bruhat intervals in W
	//
	LowerBruhatIntervalsData_t lower_intervals_data;
	
	void GetGroupAutomorphism();
	void SetGroupName();
	void GetInvolutions(std::map<CoxNbr,int> &);
	void ComputeActions(const std::map<CoxNbr,int> &);
	void RecordIndexing() const;
	void RecordIndexing_Helper(const string&, const std::vector<CoxNbr> &) const;
	void SetupBruhatOrderData();
	
	void ComputeBruhatOrder();
	void ComputeBruhatAscents();
	void ComputeBruhatAscentsByDescents();
	void ComputeBruhatIntervals_in_W();
	void ComputeBruhatAscents_Helper(const std::map<int,std::vector<int> > &, int &);
	void ComputeBruhatAscentsByDescents_Helper(Generator, int, int);
	void Compute_PsigmaPolynomial_Helper(SafePolynomial**, const CoxNbr &, const CoxNbr &) const;
	
	void AddLeadingTerms(const SafePolynomial&, const SafePolynomial&, bool, SafePolynomial &) const;
	void Subtract_Sum_z_Terms(SafePolynomial**, const Generator &, const int &, const int &, const SafePolynomial&, SafePolynomial &) const;
	void Compute_hsigma_StructureConstants_Helper(InvolIndexPol_Map**, int) const;		
	void Get_C_s_actions_Helper(InvolIndexPol_Map &, const Generator &, const int &, SafePolynomial***) const;
	void Multiply_by_C_s(InvolIndexPol_Map &, const InvolIndexPol_Map &, InvolIndexPol_Map*) const;
	void Subtract_Extra_C_z_Terms(InvolIndexPol_Map &, InvolIndexPol_Map*, const CoxNbr &, const CoxNbr &, const Generator &) const;
	void TransposeFiles(const std::vector<string>&, const std::vector<string>&) const;
	
	void PrintCoefficientStatistics(stringstream &, ComparisonData_t &) const;
	void PrintCoefficientStatistics_Helper(stringstream &, ComparisonData_t &, string, SafeCoeff &, SafePolynomial &, std::vector<int> &) const;
	
	void Test_Polynomials_Helper(ComparisonData_t &, ComparisonData_t &, ComparisonData_t &, ComparisonData_t &) const;
	void Test_StructureConstants_Helper(int, ComparisonData_t &, ComparisonData_t &, ComparisonData_t &, ComparisonData_t &) const;
	
	void Append_Group_Data(stringstream &) const;
	void Append_Comparison_Output(stringstream &, ComparisonData_t &) const;
	
	int GetInvolutiveRank(const int &) const;
	InvolBruhatIterator InitializeInterval(const int &, const int &, const Generator &) const;	
	InvolIndexPol_Map** Get_C_s_actions() const;
	SafeCoeff Sigma_Coeff_Helper(SafePolynomial**, const int &, const int &, int) const;
	
	SafePolynomial Mu_Sigma_Coeff(SafePolynomial**, const int &, const int &, const Generator &) const;
	SafePolynomial ComputeMsigmas_Helper(SafePolynomial**, const int &, const int &, const Generator&) const;
	SafePolynomial*** ComputeMsigmas() const;
	
	// These functions return the values of mu^sigma and nu^sigma as defiend in the accompanying paper.
	// The inputs are the indices of two entries in the involutions array.
	//
	SafeCoeff Mu_Sigma_Coeff(SafePolynomial** psigmas, const int &y, const int &w) const { return Sigma_Coeff_Helper(psigmas, y, w, 1); } 
	SafeCoeff Nu_Sigma_Coeff(SafePolynomial** psigmas, const int &y, const int &w) const { return Sigma_Coeff_Helper(psigmas, y, w, 2); } 
	
	// Returns list of elements y in W with y <= x and sy < y
	//
	const List<CoxNbr>& Get_LowerBruhatInterval_in_W(const Generator &s, const CoxNbr &x) const { return lower_intervals_data.d_edge[s][x]; }
	
	// Return list of numbers mu(P_{y,x}) where x varies as the  
	// list returned by Get_LowerBruhatInterval_in_W(s,x)
	//
	const List<kl::KLCoeff>& Get_LowerBruhatInterval_in_W_MuCoeffs(const Generator &s, const CoxNbr &x) const { return lower_intervals_data.d_coeff[s][x]; }	

public:
    
	TwistedInvolutionsPoset ();
	~TwistedInvolutionsPoset ();
	
	void Compute_P_Polynomials() const;
	void Compute_Psigma_Polynomials() const;
	void Compute_hsigma_StructureConstants() const;
	
	void Test_Polynomials() const;
	void Test_StructureConstants() const;
	
	void Write_hsigma_Array(int, InvolIndexPol_Map*) const; 
	
	void Transpose_h_Files() const;
	void Transpose_htilde_Files() const;
	
	void PrintPercentProgress(int &done, int total, int incr = 10) const; 
	void PrintAllProgress(const string&, int, int) const;
	
	string GetNormalForm(CoxNbr) const;
	string GetFileName(const string &, const string &) const;
	string GetFileName(const string &, const string &, int) const;
	string Get_InvolutionIndices_FileName() const;
	string Get_AllElementIndices_FileName() const;
	string Get_P_Polynomials_FileName() const;
	string Get_Psigma_Polynomials_FileName() const;
	string Get_h_StructureConstants_FileName(int) const;
	string Get_h_StructureConstants_Temp_FileName(int) const;
	string Get_htilde_StructureConstants_FileName(int) const;
	string Get_htilde_StructureConstants_Temp_FileName(int) const;
	string Get_hsigma_StructureConstants_FileName(int) const;
	string Get_Summary_FileName() const;
	
	FILE* Get_P_Polynomials_OutputFile() const;
	FILE* Get_Psigma_Polynomials_OutputFile() const;
	FILE* Get_h_StructureConstants_TempFile(int) const;
	FILE* Get_htilde_StructureConstants_TempFile(int) const;
	FILE* Get_hsigma_StructureConstants_OutputFile(int) const;
	FILE* Get_Summary_OutputFile(const char * write_type = "w")	const;
	
	bool GetLeftDescent(const int &, Generator &) const;	
	bool IsLeftDescent(const int &, const Generator &) const;
	bool Commutes(const int &, const Generator &) const;

	CoxNbr GetStarInverse(CoxNbr) const;

	// Given the index w of a twisted involution and a genrator s in S,
	// returns the index of sw (if sw=ws*) or sws* (otherwise)
	//
	int ActOnInvolution(const int & w, const Generator & s)	const { return elem_actions[w][s]; }
	
	int size()												const { return involutions.size(); }
	Length GetLength(const int &w)							const { return elem_lengths[w]; } 	
	const CoxNbr& operator[](const Ulong& j)				const { return involutions[j]; }	
	std::vector<int>& GetBruhatAscents(const int & w)		const { return bruhat_ascents[w]; }
	bool BruhatLessOrEqual(const int &y, const int &w)		const { return bruhat_order[y][w]; }
	
};



#endif
