
/* 
 * File: "CoxeterExteions_by_EM_Structs.h"
 *
 * Author: Eric Marberg
 * Version: June 2013
 * --------------------------------------------------------------
 * Header file for three structs encapsultating some auxiliary data
 * and methods used in "CoxeterExteions_by_EM_InovlPoset.cpp"
 *
 */



#ifndef __Structs__
#define __Structs__



#include <string>
#include <vector>
#include <map>
#include <set>
#include <cmath>

#include "CoxeterExtensions_by_EM_SafePolynomial.h"


using namespace std;



/*
 * Struct: ComparisonData_t
 * ---------------------------------------------------------
 * Struct associated with methods for analyzing a family of polynomials.
 * We have eight families of interest, indexed by either two or three
 * parameters:
 *
 *   P_{y,w}, Psigma_{y,w}, P^+_{y,w}, P^-_{y,w}
 *   htilde_{x,y;z}, hsigma_{x,y;z}, h^+_{x,y;z}, h^-_{x,y;z}
 *
 *
 */
struct ComparisonData_t
{
	ComparisonData_t(const string & s, char x, char y, char z='\0') 
	{ 
		tag = s;
		
		maxcoeff = SafeCoeff_MIN;
		mincoeff = SafeCoeff_MAX;
		
		polyns_decreasing = true;
		
		symbols.push_back(x);
		symbols.push_back(y);
		
		if(z != '\0') symbols.push_back(z);
	}
	
	void UpdatePolynomials(std::vector<int>*);
	void UpdateStructureConstants(int);
	void GetCoefficientStatistics();
	
	long CountProperty(bool(*test)(const SafePolynomial &));
	
	void PrintProperty(stringstream &, const string &, bool(*test)(const SafePolynomial &));
	void PrintProperty(stringstream &, const string &, bool);
	void PrintPropertiesHeader(stringstream &);
	void PrintProperties(stringstream &);
	void PrintProperties_of_Polynomials(stringstream &);
	void PrintProperties_of_StructureConstants(stringstream &);
	
	void CheckProperty(const string &, bool(*test)(const SafePolynomial &));
	void CheckProperties_of_StructureConstants();
	

	int size() { return distinct_list.size(); }
	bool isPrintingStructureConstants() { return symbols.size() == 3; }
	
	std::map<SafePolynomial,std::vector<int> > distinct_list;
	std::map<std::pair<int,int>, SafePolynomial> tmp_list;
	
	// Parameters for keeping track of largest and smallest coefficients
	std::vector<int> maxcoeff_index, mincoeff_index;
	SafePolynomial   maxcoeff_pol,   mincoeff_pol;
	SafeCoeff        maxcoeff,       mincoeff;
	
	bool polyns_decreasing;
	
	// Name of polynomials in family; e.g., "Psigma_{y,w}"
	string tag;
	
	// List of indexing parameters; typically 'y', 'w' or 'x', 'y', 'z'
	std::vector<char> symbols;
	
	std::vector<string> property_tags;
	std::map<string,bool> properties;
};





/*
 * Struct: TransposeData_t
 * ---------------------------------------------------------
 * Struct associated with the some helper methods and temporary
 * data structures used in the function TransposeFiles in 
 * "CoxeterExtensions_by_EM_InvolPoset.h".
 *
 */
struct TransposeData_t 
{
	TransposeData_t(int output_file_size) { z_lists = new std::vector<int>[output_file_size]; }
	~TransposeData_t() { delete[] z_lists; }
	
	void Setup(const string &);
	void PrintEntries(FILE*, int, int);
	
	std::set<std::pair<int,int> > yz_pairs;
	std::map<std::pair<int,int>, SafePolynomial> polynomials;
	std::vector<int>* z_lists;
};

	
	


/*
 * Struct: LowerBruhatIntervalsData_t
 * ---------------------------------------------------------
 * Struct associated with storage of lower Bruhat intervals in the 
 * whole Coxeter group W, according to a fixed descent. The code 
 * in the main methods here is copied verbatim from the
 * GraphData constructor (around line 600 in "special.cpp"), since
 * that code does what we want but can't be referenced without
 * being repeated somewhere.
 *
 */
struct LowerBruhatIntervalsData_t
{	
	void AllocateLists(const Ulong &, const Rank&);		
	void FillLists(const wgraph::WGraph &);	
	
	List<List<List<CoxNbr> > > d_edge;
    List<List<List<kl::KLCoeff> > > d_coeff;	
};





#endif